#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <stdio.h>

#ifndef USE_DX9
#include <d3dx8.h>
#else
#include <d3dx9.h>
#endif

HWND hWnd;

#ifndef USE_DX9
LPDIRECT3D8 p_d3d;
LPDIRECT3DDEVICE8 p_d3d_Device;
IDirect3DTexture8* tb;
#define def_tf D3DX_DEFAULT
#else
LPDIRECT3D9 p_d3d;
LPDIRECT3DDEVICE9 p_d3d_Device;
IDirect3DTexture9* tb;
#define def_tf D3DX_DEFAULT_NONPOW2
#endif

D3DDISPLAYMODE d3ddm;
D3DPRESENT_PARAMETERS d3dpp;
D3DXIMAGE_INFO ii;
DWORD tsize;

#define def_pool D3DPOOL_MANAGED  // D3DPOOL_DEFAULT

#define _RELEASE_(p) { if(p) { (p)->Release(); (p)=NULL; };};
#define _DELETE_(p)  { if(p) { delete (p);     (p)=NULL; };};

char mbuf[255];

#if (D3D_SDK_VERSION == 120)
HRESULT D3DXGetImageInfoFromFile(
																 LPCSTR pSrcFile,
																 D3DXIMAGE_INFO *pSrcInfo
																 )
{
	LPDIRECT3DTEXTURE8 tmpTB;
	
	HRESULT hr = D3DXCreateTextureFromFileEx(p_d3d_Device,
		pSrcFile,def_tf,def_tf,1,0, // def_usage = 0, mipmaps = 1
		D3DFMT_UNKNOWN,def_pool,D3DX_FILTER_POINT,D3DX_FILTER_NONE,0,pSrcInfo,NULL,
		&tmpTB);
	_RELEASE_( tmpTB ); _DELETE_( tmpTB );
	
	return hr;
}
#endif

_D3DDEVTYPE devtype[3] = {D3DDEVTYPE_HAL,D3DDEVTYPE_SW,D3DDEVTYPE_REF};

typedef struct {
   char  idlength;
   char  colourmaptype;
   char  datatypecode;
   short int colourmaporigin;
   short int colourmaplength;
   char  colourmapdepth;
   short int x_origin;
   short int y_origin;
   short width;
   short height;
   char  bitsperpixel;
   char  imagedescriptor;
} TGAHEADER;

int main(int argc, char* argv[])
{
	TGAHEADER th;
	unsigned int e;
	unsigned char vflag = 0, mflag = 0, a = 0;
	FILE * inputFile, * outputFile;
	char fileName[_MAX_PATH], ofileName[_MAX_PATH], tmpName[_MAX_PATH];
#ifdef USE_DX9
	HMODULE hDll;
	FARPROC p;
#endif	

 	printf("NF surface texture converter [201203xx]\n");
	
	if (argc < 2) 
  {
		printf("usage: surfconf < srcfile.ext > [ destfile.ext ] [-flag1] [-flag2]\n");
		printf(" if destfile omitted it will be saved as destfile.ext, same format as input\n");
		printf(" -v : verbose\n");
		printf("  flags can be :\n");
		printf(" -invrgb : it inverts order (RGBA <-> BGRA)\n");
		printf(" -invnrm : it inverts order (RGBA <-> ABGR, Nxyzw <-> Nwzyx)\n");
		return 1;
	}
	
  mflag = 0;
	for (a = 2; a < argc;a++) 
	{
		strcpy(tmpName,argv[a]);
		if (strstr(tmpName,"-v")) vflag = 1, printf("* verbose ON\n");;
		if (strstr(tmpName,"-invrgb")) mflag = 1, printf("* mode INVRGB\n");
		if (strstr(tmpName,"-invnrm")) mflag = 2, printf("* mode INVNRM\n");
	}
	
	if (strstr(argv[2],"-")) strcpy((char*)ofileName,"outfile.ext"); 
	else strcpy((char*)ofileName,argv[2]);

	unsigned short olen;		
	olen = strlen(ofileName) - 4; // i.e. "????.ext"

	strcpy((char*)fileName,argv[1]);
	inputFile = fopen(fileName,"rb");
	if (!inputFile) { printf("'%s' not found.\n",fileName); goto exit0; }
	fseek(inputFile,0,SEEK_END);
  tsize = ftell(inputFile);
	fclose(inputFile);
	
	hWnd = CreateWindow("STATIC","default",NULL,0,0,320,200,NULL,NULL,NULL,NULL);
	ShowWindow(hWnd,SW_SHOWNORMAL);
	
	p_d3d_Device = NULL;
	p_d3d = NULL;
	
#ifndef USE_DX9	
	p_d3d = Direct3DCreate8 (120); // D3D_SDK_VERSION
	if  (p_d3d == NULL) { printf("! Direct3DCreate8() == NULL\n"); goto exit0; }
#else
	p_d3d = Direct3DCreate9 (D3D_SDK_VERSION);
	if  (p_d3d == NULL) { printf("! Direct3DCreate9() == NULL\n"); goto exit0; }
	
	hDll = LoadLibrary("rgb9rast.dll");
	if (!hDll) hDll = LoadLibrary("rgb9rast_2.dll");
	if (!hDll) hDll = LoadLibrary("rgb9rast_1.dll");
	p = GetProcAddress(hDll, "D3D9GetSWInfo");
	p_d3d->RegisterSoftwareDevice(p);
#endif
	
	ZeroMemory (&d3ddm, sizeof(d3ddm));
	
	p_d3d->GetAdapterDisplayMode (D3DADAPTER_DEFAULT, &d3ddm);
	ZeroMemory (&d3dpp, sizeof(d3dpp));
	d3dpp.Windowed = true;
	d3dpp.MultiSampleType = D3DMULTISAMPLE_NONE;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = d3ddm.Format;
	d3dpp.BackBufferWidth = d3ddm.Width;
	d3dpp.BackBufferHeight = d3ddm.Height;
	d3dpp.BackBufferCount = 1;
	d3dpp.hDeviceWindow = hWnd;
	d3dpp.EnableAutoDepthStencil = true;
	d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	
	for (a = 0; a < 3; a++)
	{
	e = p_d3d->CreateDevice(D3DADAPTER_DEFAULT, devtype[a],
				hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &p_d3d_Device);
		if (e == D3D_OK) { if (vflag)  { printf("p_d3d->CreateDevice() OK\n"); } break; }
		if (e != D3D_OK) printf("! p_d3d->CreateDevice(... devtype = %u, ...) error code:0x%08X\n",a,e);
	}	
		printf(" converting '%s' -> ",fileName);
		
		////////////////////////////////////////////////////////////////////////////		
		
		e = D3DXGetImageInfoFromFile(fileName,&ii);
		if (vflag) if (!strstr(&ofileName[olen],".dds")) ii.MipLevels = 1, printf("\n* mipmaps DISABLED\n");
		if (vflag) if (e == D3D_OK) printf("\nD3DXGetImageInfoFromFile() OK\n");
		if (e != D3D_OK)
			printf("! D3DXGetImageInfoFromFile() error code:0x%08Xn",e);
		
		if (vflag) printf(" -> texture format w:%u, h:%u, mips:%u, ifformat:%02X, format:%08X\n",
			ii.Width,ii.Height,ii.MipLevels,ii.ImageFileFormat,ii.Format);
		
		e = D3DXCreateTextureFromFileEx(p_d3d_Device, fileName,
			def_tf,def_tf,ii.MipLevels,0, // def_usage = 0
			D3DFMT_A8R8G8B8,def_pool,D3DX_FILTER_POINT,D3DX_FILTER_NONE,0,&ii,NULL, &tb);
		
		if (e == D3D_OK)
		{
			if (vflag) printf("D3DXCreateTextureFromFileEx() OK\n");
			////////////////////////////////////////////////////////////////////////////		
			D3DLOCKED_RECT LockedRect;
			DWORD * pDword, pTemp;
			if (mflag)
			for (unsigned int ml = 0; ml < ii.MipLevels; ml++)
			{
				tb->LockRect(ml,&LockedRect,NULL,0);
				pDword = (DWORD *)LockedRect.pBits;
				
				for (unsigned int ix = 0; ix < ii.Width; ix ++)
					for (unsigned int iy = 0; iy < ii.Height; iy ++)
					{
						// aabbccdd == RGBA -> ARGB -> RABG ( QWVU -> UQWV -> QUVW )
						// aa = hmap (R,Q)
						// bb = nrm, (G,W,N)
						// cc = vrt, (B,V,B)
						// dd = hrz, (A,U,T)
						
						if (mflag == 1) pTemp = ((*pDword & 0xFF00FF00) | (*pDword & 0x000000FF) << 16) | ((*pDword & 0x00FF0000) >> 16);
						else if (mflag == 2) pTemp = ((*pDword & 0x00FF00FF) << 8) | ((*pDword & 0xFF00FF00) >> 8);
						*pDword = pTemp;
						//	*pDword = 0xFF8080FF;
						pDword++;
					}
					tb->UnlockRect(ml);
					if (ii.MipLevels == 1) break;
					ii.Width >>= 1; ii.Height >>= 1; // next mip
			}
			////////////////////////////////////////////////////////////////////////////		
		}
		else
		{
			printf("! D3DXCreateTextureFromFileEx() error code:0x%08X\n",e);
		}

		if (strstr(&ofileName[olen],".bmp")) ii.ImageFileFormat = D3DXIFF_BMP;
		else if (strstr(&ofileName[olen],".jpg")) ii.ImageFileFormat = D3DXIFF_JPG;
		else if (strstr(&ofileName[olen],".tga")) ii.ImageFileFormat = D3DXIFF_TGA;
		else if (strstr(&ofileName[olen],".png")) ii.ImageFileFormat = D3DXIFF_PNG;
		else if (strstr(&ofileName[olen],".dds")) ii.ImageFileFormat = D3DXIFF_DDS;
		else if (strstr(&ofileName[olen],".ppm")) ii.ImageFileFormat = D3DXIFF_PPM;
		else if (strstr(&ofileName[olen],".dib")) ii.ImageFileFormat = D3DXIFF_DIB;
#ifdef USE_DX9
		else if (strstr(&ofileName[olen],".hdr")) ii.ImageFileFormat = D3DXIFF_HDR;
		else if (strstr(&ofileName[olen],".pfm")) ii.ImageFileFormat = D3DXIFF_PFM;
#endif
		
		printf("'%s'", ofileName);
		if (ii.ImageFileFormat != D3DXIFF_TGA)
		e = D3DXSaveTextureToFile(ofileName,ii.ImageFileFormat,tb,NULL);
		else
		for (;;) {
		th.idlength = th.colourmaptype = 
		th.colourmaporigin = th.colourmaplength = th.colourmapdepth =
		th.x_origin = th.y_origin = th.imagedescriptor = 0;
		th.datatypecode = 2; // RGB
   	th.width = ii.Width;
   	th.height = ii.Height;
		th.bitsperpixel = 24;
		outputFile = fopen(ofileName,"wb");
		if (!outputFile) { e = 0xDEADC0DE; break; }
		fwrite(&th,sizeof(TGAHEADER),1,outputFile);
			////////////////////////////////////////////////////////////////////////////		
			D3DLOCKED_RECT LockedRect;
			DWORD * pDword;
				tb->LockRect(0,&LockedRect,NULL,0);
				pDword = (DWORD *)LockedRect.pBits;
				
				for (unsigned int ix = 0; ix < ii.Width; ix ++)
					for (unsigned int iy = 0; iy < ii.Height; iy ++)
					{
						pDword++;
						fwrite(pDword,3,1,outputFile);
					}
					tb->UnlockRect(0);
			////////////////////////////////////////////////////////////////////////////		
    fclose(outputFile);
		e = D3D_OK; 
		break;
		}

		if (vflag) if (e == D3D_OK) printf("\nD3DXSaveTextureToFile() OK\n");
		if (e != D3D_OK) printf("\n! D3DXSaveTextureToFile() error code:0x%08X\n",e);
		printf(" OK\n");
		
		if (tb != NULL) {
			_try { _RELEASE_(tb); _DELETE_(tb); }
			_except (1) { printf("exception while cleaning '%s'",mbuf); } }
		
		if (ii.ImageFileFormat == D3DXIFF_DDS)
		{
			printf(" recompress '%s' ...", ofileName);
			e = D3DXCreateTextureFromFileEx(p_d3d_Device,
				ofileName,def_tf,def_tf,ii.MipLevels,0, // def_usage = 0
				ii.Format,def_pool,D3DX_FILTER_POINT,D3DX_FILTER_NONE,0,&ii,NULL,
				&tb);
			
			if (e == D3D_OK)
			{
				if (vflag) printf("\nD3DXCreateTextureFromFileEx() OK\n");
				////////////////////////////////////////////////////////////////////////////		
				e = D3DXSaveTextureToFile(ofileName,ii.ImageFileFormat,tb,NULL);
				if (vflag) if (e == D3D_OK) printf("D3DXSaveTextureToFile() OK\n");
				if (e != D3D_OK) printf("\n! D3DXSaveTextureToFile() error code:0x%08X\n",e);
				printf(" OK\n");
				
				if (tb != NULL) {
					_try { _RELEASE_(tb); _DELETE_(tb); }
					_except (1) { printf("exception while cleaning '%s'",mbuf); } }
				////////////////////////////////////////////////////////////////////////////		
			}
			else
			{
				printf("\n! D3DXCreateTextureFromFileEx() error code:0x%08X\n",e);
			}
			
		}
		////////////////////////////////////////////////////////////////////////////		
		
		if (vflag) printf("processing DONE\n");
		
		_RELEASE_(p_d3d_Device); _DELETE_(p_d3d_Device);
		_RELEASE_(p_d3d); _DELETE_(p_d3d);
		DestroyWindow(hWnd);
		if (vflag) printf("return OK\n");
exit0:
		return 0;
}